#include <Arduino.h>
#include <Wire.h>
#include "MPU9250.h"
#include "util.h"

MPU9250::MPU9250() {	
	// valid for accelerometer full scale = +/- 2G
    axBias_ = 105;
    ayBias_ = -273;
    azBias_ = -508;
    aScale_ = 1.0f/MPU9250_2G_SENSITIVITY; // accelerometer values in milli-Gs
    
    // valid for gyro full scale = 500dps
    gxBias_ = 42; 
    gyBias_ = -18; 
    gzBias_ = 73; 
    gScale_  = 1.0f/MPU9250_500DPS_SENSITIVITY; // gyroscope values in deg/second
	}

void MPU9250::GetAccelGyroData(float* pAccelData, float* pGyroData) {
	uint8_t buf[14];
	int16_t raw[3];
	ReadBytes(MPU9250_I2C_ADDRESS, ACCEL_XOUT_H, 14, buf);
	raw[0] = (int16_t)(((uint16_t)buf[0] << 8) | (uint16_t)buf[1]);
	raw[1] = (int16_t)(((uint16_t)buf[2] << 8) | (uint16_t)buf[3]);
	raw[2] = (int16_t)(((uint16_t)buf[4] << 8) | (uint16_t)buf[5]);	
	pAccelData[0] = (float)(raw[0] - axBias_) * aScale_;
	pAccelData[1] = (float)(raw[1] - ayBias_) * aScale_;
	pAccelData[2] = (float)(raw[2] - azBias_) * aScale_;
	raw[0] = (int16_t)(((uint16_t)buf[8] << 8) | (uint16_t)buf[9]);
	raw[1] = (int16_t)(((uint16_t)buf[10] << 8) | (uint16_t)buf[11]);
	raw[2] = (int16_t)(((uint16_t)buf[12] << 8) | (uint16_t)buf[13]);	
    pGyroData[0] = (float)(raw[0] - gxBias_) * gScale_;
    pGyroData[1] = (float)(raw[1] - gyBias_) * gScale_;
    pGyroData[2] = (float)(raw[2] - gzBias_) * gScale_;
	}

	
int MPU9250::CheckID(void) {
	uint8_t whoami = ReadByte(MPU9250_I2C_ADDRESS, WHO_AM_I_MPU9250);
	return (( whoami == 0x71) ? 1 : 0);
	}
	
void MPU9250::SetCalibrationParams(NVD* pNVD) {
	axBias_ = pNVD->params.axBias;
	ayBias_ = pNVD->params.ayBias;
	azBias_ = pNVD->params.azBias;
	gxBias_ = pNVD->params.gxBias;
	gyBias_ = pNVD->params.gyBias;
	gzBias_ = pNVD->params.gzBias;
	Serial.println("Using saved calibration parameters :");
	char tmpbuffer[50];
	sprintf(tmpbuffer,"Accel : axBias %d, ayBias %d, azBias %d",axBias_, ayBias_, azBias_);
	Serial.println(tmpbuffer);
	sprintf(tmpbuffer,"Gyro : gxBias %d, gyBias %d, gzBias %d",gxBias_, gyBias_, gzBias_);
	Serial.println(tmpbuffer);
	}

void MPU9250::Sleep(void) {
	WriteByte(MPU9250_I2C_ADDRESS, PWR_MGMT_1, 0x40);
	}

void MPU9250::ConfigAccelGyro(void) {
	// reset MPU9250, all registers to default settings
	WriteByte(MPU9250_I2C_ADDRESS, PWR_MGMT_1, 0x80);
	delay(100); // Wait after reset
	// as per datasheet all registers are reset to 0 except WHOAMI and PWR_MGMT_1, 
	// so we assume reserved bits are 0
	// select best available clock source 
	WriteByte(MPU9250_I2C_ADDRESS, PWR_MGMT_1, 0x01);
	delay(200);

	// fsync disabled, gyro bandwidth = 41Hz (with GYRO_CONFIG:fchoice = 11) 
	WriteByte(MPU9250_I2C_ADDRESS, CONFIG, 0x03);

	// output data rate = 1000Hz/5 = 200Hz
	WriteByte(MPU9250_I2C_ADDRESS, SMPLRT_DIV, 0x04);

	// set gyro FS = 500dps, Fchoice = b11 (inverse of bits [1:0] = 00)
	WriteByte(MPU9250_I2C_ADDRESS, GYRO_CONFIG, 0x08 );

	// Set accelerometer FS = +/-2G 
	// for aFS = 2g, bits[4:3] = 00 
	WriteByte(MPU9250_I2C_ADDRESS, ACCEL_CONFIG, 0x00);

	// set accelerometer BW = 41Hz
	// accel_fchoice = 1 (inverse of bit 3), a_dlpf_cfg bits[2:0] = 011
	WriteByte(MPU9250_I2C_ADDRESS, ACCEL_CONFIG2, 0x03);

	// interrupt is active high, push-pull, 50uS pulse
	WriteByte(MPU9250_I2C_ADDRESS, INT_PIN_CFG, 0x10);
	// Enable data ready interrupt on INT pin
	WriteByte(MPU9250_I2C_ADDRESS, INT_ENABLE, 0x01);
	delay(100);
	}


// place unit so that the sensor board accelerometer +ve z axis points 
// vertically downwards. This is where the sensor z axis is experiencing a static 
// acceleration of 1g. In this orientation the ax and ay values are 
// the offsets for a 0g environment. 
// Repeat this calibration a few times with the debug serial monitor to check the 
// consistency of the calibration offsets. The board MUST be in a 1g static acceleration 
// environment for this calibration, i.e. at rest, no vibrations etc.

#define ACCEL_NUM_AVG_SAMPLES	10
#define ACCEL_NUM_CALIB_TRIES	200

void MPU9250::CalibrateAccel(){
	uint8_t buf[6];
	int16_t ax,ay,az,ax0g,ay0g;
	int16_t az1g = 0;
	int32_t axAccum, ayAccum, azAccum;
	Serial.println("");
	char tmpbuffer[50];

	for (int cnt = 0; cnt < ACCEL_NUM_CALIB_TRIES; cnt++) {
		axAccum = ayAccum = azAccum = 0;
		for (int inx = 0; inx < ACCEL_NUM_AVG_SAMPLES; inx++){
			ReadBytes(MPU9250_I2C_ADDRESS, ACCEL_XOUT_H, 6, buf);
			ax = (int16_t)(((uint16_t)buf[0] << 8) | (uint16_t)buf[1]);
			ay = (int16_t)(((uint16_t)buf[2] << 8) | (uint16_t)buf[3]);
			az = (int16_t)(((uint16_t)buf[4] << 8) | (uint16_t)buf[5]);	
			axAccum += (int32_t) ax;
			ayAccum += (int32_t) ay;
			azAccum += (int32_t) az;
			delay(5);
			}
		ax = (int16_t)(axAccum / ACCEL_NUM_AVG_SAMPLES);
		ay = (int16_t)(ayAccum / ACCEL_NUM_AVG_SAMPLES);
		az = (int16_t)(azAccum / ACCEL_NUM_AVG_SAMPLES);
		if (az > az1g) {
			az1g = az;
			ax0g = ax;
			ay0g = ay;
			}
		sprintf(tmpbuffer,"%03d/%3d : ax %d  ay %d  az %d\r\n",cnt, ACCEL_NUM_CALIB_TRIES, ax, ay, az);
		Serial.println(tmpbuffer);
		}

	axBias_ = ax0g;
	ayBias_ = ay0g;
    azBias_ = az1g - (int16_t)(1000.0f*MPU9250_2G_SENSITIVITY);

    sprintf(tmpbuffer,"axBias = %d", (int)axBias_);
	Serial.println(tmpbuffer);
    sprintf(tmpbuffer,"ayBias = %d", (int)ayBias_);
	Serial.println(tmpbuffer);
    sprintf(tmpbuffer,"azBias = %d", (int)azBias_);
	Serial.println(tmpbuffer);
	}

#define GYRO_MAX_EXPECTED_OFFSET_500DPS	100
#define GYRO_NUM_CALIB_SAMPLES			50
	
void MPU9250::CalibrateGyro(void){
	uint8_t buf[6];
	int16_t gx,gy,gz;
	int32_t gxAccum, gyAccum, gzAccum;
	int foundBadData;
	int numTries = 1;
	do {
		delay(500);
		foundBadData = 0;
		gxAccum = gyAccum = gzAccum = 0;
		for (int inx = 0; inx < GYRO_NUM_CALIB_SAMPLES; inx++){
			ReadBytes(MPU9250_I2C_ADDRESS, GYRO_XOUT_H, 6, buf);
			gx = (int16_t)(((uint16_t)buf[0] << 8) | (uint16_t)buf[1]);
			gy = (int16_t)(((uint16_t)buf[2] << 8) | (uint16_t)buf[3]);
			gz = (int16_t)(((uint16_t)buf[4] << 8) | (uint16_t)buf[5]);	
			if ((ABS(gx) > GYRO_MAX_EXPECTED_OFFSET_500DPS) || (ABS(gy) > GYRO_MAX_EXPECTED_OFFSET_500DPS) || (ABS(gz) > GYRO_MAX_EXPECTED_OFFSET_500DPS)) {
				foundBadData = 1;
				break;
				}  
			gxAccum  += (int32_t) gx;
			gyAccum  += (int32_t) gy;
			gzAccum  += (int32_t) gz;
			delay(5);
			}
		} while (foundBadData && (++numTries < 10));

    if (!foundBadData) {		
		gxBias_ =  (int16_t)( gxAccum / GYRO_NUM_CALIB_SAMPLES);
		gyBias_ =  (int16_t)( gyAccum / GYRO_NUM_CALIB_SAMPLES);
		gzBias_ =  (int16_t)( gzAccum / GYRO_NUM_CALIB_SAMPLES);
		}

    char tmpbuffer[50];
	sprintf(tmpbuffer,"Num Tries = %d\r\n",numTries);
	Serial.println(tmpbuffer);
	sprintf(tmpbuffer,"gxBias = %d\r\n",gxBias_);
    Serial.println(tmpbuffer);
    sprintf(tmpbuffer,"gyBias = %d\r\n",gyBias_);
	Serial.println(tmpbuffer);
	sprintf(tmpbuffer,"gzBias = %d\r\n",gzBias_);
	Serial.println(tmpbuffer);
	}


void MPU9250::WriteByte(uint8_t deviceAddress, uint8_t registerAddress, uint8_t d) {
  Wire.beginTransmission(deviceAddress);  
  Wire.write(registerAddress);
  Wire.write(d);           
  Wire.endTransmission();     
}

uint8_t MPU9250::ReadByte(uint8_t deviceAddress, uint8_t registerAddress){
  uint8_t d; 
  Wire.beginTransmission(deviceAddress);
  Wire.write(registerAddress);
  Wire.endTransmission(false); // restart
  Wire.requestFrom(deviceAddress, (uint8_t) 1);
  d = Wire.read();
  return d;
}

int MPU9250::ReadBytes(uint8_t deviceAddress, uint8_t registerAddress,
                        uint8_t count, uint8_t * dest) {
  Wire.beginTransmission(deviceAddress);
  Wire.write(registerAddress);
  Wire.endTransmission(false); // restart
  int cnt = 0;
  Wire.requestFrom(deviceAddress, count);
  while (Wire.available())  {
    dest[cnt++] = Wire.read();
	}
  return cnt;
}

/*
// multiplicative conversion constants 
#define DegToRad 0.017453292F
#define RadToDeg 57.295779F
// global variables 
float Psi, The, Phi; // yaw, pitch, roll angles in deg 
float Vx, Vy, Vz; // hard iron calibration coefficients 
// NED tilt-compensated e-Compass function 
void eCompass(float Bx, float By, float Bz, float Gx, float Gy, float Gz)
{
float sinAngle, cosAngle; // sine and cosine 
float fBfx, fBfy, fBfz; // calibrated mag data in uT after tilt correction 
// subtract off the hard iron interference computed using equation 9 
Bx -= Vx;
By -= Vy;
Bz -= Vz;
// calculate roll angle Phi (-180deg, 180deg) and sin, cos 
Phi = atan2(Gy, Gz) * RadToDeg; // Equation 2 
sinAngle = sin(Phi * DegToRad); // sin(Phi) 
cosAngle = cos(Phi * DegToRad); // cos(Phi) 
// de-rotate by roll angle Phi 
Bfy = By * cosAngle - Bz * sinAngle; // Equation 5 y component 
Bz = By * sinAngle + Bz * cosAngle; // Bz=(By-Vy).sin(Phi)+(Bz-Vz).cos(Phi) 
Gz = Gy * sinAngle + Gz * cosAngle; // Gz=Gy.sin(Phi)+Gz.cos(Phi) 
// calculate pitch angle Theta (-90deg, 90deg) and sin, cos
The = atan(-Gx / Gz) * RadToDeg; // Equation 3 
sinAngle = sin(The * DegToRad); // sin(Theta) 
cosAngle = cos(The * DegToRad); // cos(Theta) 
// de-rotate by pitch angle Theta 
Bfx = Bx * cosAngle + Bz * sinAngle; // Equation 5 x component 
Bfz = -Bx * sinAngle + Bz * cosAngle; // Equation 5 z component 
// calculate yaw = ecompass angle psi (-180deg, 180deg) 
Psi = atan2(-Bfy, Bfx) * RadToDeg; // Equation 7 
return;
}
*/